from enum import IntEnum


class Manufacturer(IntEnum):
    Reserved = 0x000
    PEHA = 0x001
    THERMOKON = 0x002
    SERVODAN = 0x003
    ECHOFLEX_SOLUTIONS = 0x004
    OMNIO_AG = 0x005
    AWAG_ELEKTROTECHNIK_AG = 0x005
    HARDMEIER_ELECTRONICS = 0x006
    REGULVAR_INC = 0x007
    AD_HOC_ELECTRONICS = 0x008
    DISTECH_CONTROLS = 0x009
    KIEBACK_AND_PETER = 0x00A
    ENOCEAN_GMBH = 0x00B
    PROBARE = 0x00C
    VICOS_GMBH = 0x00C
    ELTAKO = 0x00D
    LEVITON = 0x00E
    HONEYWELL = 0x00F
    SPARTAN_PERIPHERAL_DEVICES = 0x010
    SIEMENS = 0x011
    T_MAC = 0x012
    RELIABLE_CONTROLS_CORPORATION = 0x013
    ELSNER_ELEKTRONIK_GMBH = 0x014
    DIEHL_CONTROLS = 0x015
    BSC_COMPUTER = 0x016
    S_AND_S_REGELTECHNIK_GMBH = 0x017
    ZENO_CONTROLS = 0x018
    MASCO_CORPORATION = 0x018
    INTESIS_SOFTWARE_SL = 0x019
    VIESSMANN = 0x01A
    LUTUO_TECHNOLOGY = 0x01B
    CAN2GO = 0x01C
    SAUTER = 0x01D
    BOOT_UP = 0x01E
    OSRAM_SYLVANIA = 0x01F
    UNOTECH = 0x020
    DELTA_CONTROLS_INC = 0x21
    UNITRONIC_AG = 0x022
    NANOSENSE = 0x023
    THE_S4_GROUP = 0x024
    MSR_SOLUTIONS = 0x025
    VEISSMANN_HAUSATOMATION_GMBH = 0x025
    GE = 0x26
    MAICO = 0x027
    RUSKIN_COMPANY = 0x28
    MAGNUM_ENERGY_SOLUTIONS = 0x29
    KMC_CONTROLS = 0x02A
    ECOLOGIX_CONTROLS = 0x02B
    TRIO_2_SYS = 0x2C
    AFRISO_EURO_INDEX = 0x02D
    WALDMANN_GMBH = 0x02E
    NEC_PLATFORMS_LTD = 0x030
    ITEC_CORPORATION = 0x031
    SIMICX_CO_LTD = 0x32
    PERMUNDO_GMBH = 0x33
    EUROTRONIC_TECHNOLOGY_GMBH = 0x34
    ART_JAPAN_CO_LTD = 0x35
    TIANSU_AUTOMATION_CONTROL_SYSTE_CO_LTD = 0x36
    WEINZIERL_ENGINEERING_GMBH = 0x37
    GRUPPO_GIORDANO_IDEA_SPA = 0x38
    ALPHAEOS_AG = 0x39
    TAG_TECHNOLOGIES = 0x3A
    WATTSTOPPER = 0x3B
    PRESSAC_COMMUNICATIONS_LTD = 0x3C
    GIGA_CONCEPT = 0x3E
    SENSORTEC = 0x3F
    JAEGER_DIREKT = 0x40
    AIR_SYSTEM_COMPONENTS_INC = 0x41
    ERMINE_CORP = 0x042
    SODA_GMBH = 0x043
    EKE_AUTOMATION = 0x044
    HOLTER_REGELARMUTREN = 0x045
    ID_RF = 0x046
    DEUTA_CONTROLS_GMBH = 0x047
    EWATTCH = 0x048
    MICROPELT = 0x049
    CALEFFI_SPA = 0x04A
    DIGITAL_CONCEPTS = 0x04B
    EMERSON_CLIMATE_TECHNOLOGIES = 0x04C
    ADEE_ELECTRONIC = 0x04D
    ALTECON = 0x04E
    NANJING_PUTIAN_TELECOMMUNICATIONS = 0x04F
    TERRALUX = 0x050
    MENRED = 0x051
    IEXERGY_GMBH = 0x052
    OVENTROP_GMBH = 0x053
    BUILDING_AUTOMATION_PRODUCTS_INC = 0x054
    FUNCTIONAL_DEVICES_INC = 0x055
    OGGA = 0x056
    ITHO_DAALDEROP = 0x057
    RESOL = 0x058
    ADVANCED_DEVICES = 0x059
    AUTANI_LCC = 0x05A
    DR_RIEDEL_GMBH = 0x05B
    HOPPE_HOLDING_AG = 0x05C
    SIEGENIA_AUBI_KG = 0x05D
    ADEO_SERVICES = 0x05E
    EIMSIG_EFP_GMBH = 0x05F
    VIMAR_SPA = 0x060
    GLEN_DIMLAX_GMBH = 0x061
    PMDM_GMBH = 0x062
    HUBBEL_LIGHTNING = 0x063
    DEBFLEX = 0x64
    PERIFACTORY_SENSORSYSTEMS = 0x65
    WATTY_CORP = 0x66
    WAGO_KONTAKTTECHNIK = 0x67
    KESSEL = 0x68
    AUG_WINKHAUS = 0x69
    DECELECT = 0x6A
    MST_INDUSTRIES = 0x6B
    BECKER_ANTRIEBE = 0x6C
    NEXELEC = 0x6D
    WIELAND_ELECTRIC = 0x6E
    AVIDSEN = 0x6F
    CWS_BOCO_INTERNATIONAL = 0x70
    ROTO_FRANK = 0x71
    ALM_CONTORLS = 0x072
    TOMMASO_TECHNOLOGIES = 0x073
    REHAU = 0x074
    MULTI_USER_MANUFACTURER = 0x7FF
